######################################################################
#
# File: test/unit/_cli/fixtures/module_loading_b2sdk.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################

# This is a helper module for test_autocomplete_cache.py

from b2sdk.v2 import B2Api  # noqa


def function():
    pass



######################################################################
#
# File: b2/_internal/_utils/python_compat.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
"""
Utilities for compatibility with older Python versions.
"""
import functools
import shlex
import sys

if sys.version_info < (3, 9):

    def removeprefix(s: str, prefix: str) -> str:
        return s[len(prefix):] if s.startswith(prefix) else s

else:
    removeprefix = str.removeprefix

if sys.version_info < (3, 8):

    class singledispatchmethod:
        """
        singledispatchmethod backport for Python 3.7.

        There are no guarantees for its completeness.
        """

        def __init__(self, method):
            self.dispatcher = functools.singledispatch(method)
            self.method = method

        def register(self, cls, method=None):
            return self.dispatcher.register(cls, func=method)

        def __get__(self, obj, cls):
            @functools.wraps(self.method)
            def method_wrapper(arg, *args, **kwargs):
                method_desc = self.dispatcher.dispatch(arg.__class__)
                return method_desc.__get__(obj, cls)(arg, *args, **kwargs)

            method_wrapper.register = self.register
            return method_wrapper

    def shlex_join(split_command):
        return ' '.join(shlex.quote(arg) for arg in split_command)
else:
    singledispatchmethod = functools.singledispatchmethod
    shlex_join = shlex.join



######################################################################
#
# File: test/unit/helpers.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import concurrent.futures
import sys


class RunOrDieExecutor(concurrent.futures.ThreadPoolExecutor):
    """
    Deadly ThreadPoolExecutor, which ensures all task are quickly closed before exiting.

    Only really usable in tests.
    """

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown(wait=False, cancel_futures=True)
        return super().__exit__(exc_type, exc_val, exc_tb)

    if sys.version_info < (3, 9):  # shutdown(cancel_futures=True) is Python 3.9+

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._futures = []

        def shutdown(self, wait=True, cancel_futures=False):
            if cancel_futures:
                for future in self._futures:
                    future.cancel()
            super().shutdown(wait=wait)

        def submit(self, *args, **kwargs):
            future = super().submit(*args, **kwargs)
            self._futures.append(future)
            return future



######################################################################
#
# File: test/unit/console_tool/test_help.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import pytest


@pytest.mark.parametrize(
    "flag, included, excluded",
    [
        # --help shouldn't show deprecated commands
        (
            "--help",
            [" b2 download-file ", "-h", "--help-all"],
            [" download-file-by-name ", "(DEPRECATED)"],
        ),
        # --help-all should show deprecated commands, but marked as deprecated
        (
            "--help-all",
            ["(DEPRECATED) b2 download-file-by-name ", "-h", "--help-all"],
            [],
        ),
    ],
)
def test_help(b2_cli, flag, included, excluded, capsys):
    b2_cli.run([flag], expected_stdout=None)

    out = capsys.readouterr().out

    found = set()
    for i in included:
        if i in out:
            found.add(i)
    for e in excluded:
        if e in out:
            found.add(e)
    assert found.issuperset(included), f"expected {included!r} in {out!r}"
    assert found.isdisjoint(excluded), f"expected {excluded!r} not in {out!r}"



######################################################################
#
# File: test/unit/console_tool/test_upload_file.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import os
from test.helpers import skip_on_windows

import b2


def test_upload_file__file_info_src_last_modified_millis_and_headers(b2_cli, bucket, tmpdir):
    """Test upload_file supports manually specifying file info src_last_modified_millis"""
    filename = 'file1.txt'
    content = 'hello world'
    local_file1 = tmpdir.join('file1.txt')
    local_file1.write(content)

    expected_json = {
        "action": "upload",
        "contentSha1": "2aae6c35c94fcfb415dbe95f408b9ce91ee846ed",
        "fileInfo":
            {
                "b2-cache-control": "max-age=3600",
                "b2-expires": "Thu, 01 Dec 2050 16:00:00 GMT",
                "b2-content-language": "en",
                "b2-content-disposition": "attachment",
                "b2-content-encoding": "gzip",
                "src_last_modified_millis": "1"
            },
        "fileName": filename,
        "size": len(content),
    }
    b2_cli.run(
        [
            'upload-file', '--no-progress', '--info=src_last_modified_millis=1', 'my-bucket',
            '--cache-control', 'max-age=3600', '--expires', 'Thu, 01 Dec 2050 16:00:00 GMT',
            '--content-language', 'en', '--content-disposition', 'attachment', '--content-encoding',
            'gzip',
            str(local_file1), 'file1.txt'
        ],
        expected_json_in_stdout=expected_json,
        remove_version=True,
    )


@skip_on_windows
def test_upload_file__named_pipe(b2_cli, bucket, tmpdir, bg_executor):
    """Test upload_file supports named pipes"""
    filename = 'named_pipe.txt'
    content = 'hello world'
    local_file1 = tmpdir.join('file1.txt')
    os.mkfifo(str(local_file1))
    writer = bg_executor.submit(
        local_file1.write, content
    )  # writer will block until content is read

    expected_stdout = f'URL by file name: http://download.example.com/file/my-bucket/{filename}'
    expected_json = {
        "action": "upload",
        "contentSha1": "2aae6c35c94fcfb415dbe95f408b9ce91ee846ed",
        "contentType": "b2/x-auto",
        "fileName": filename,
        "size": len(content),
    }
    b2_cli.run(
        ['upload-file', '--no-progress', 'my-bucket',
         str(local_file1), filename],
        expected_json_in_stdout=expected_json,
        remove_version=True,
        expected_part_of_stdout=expected_stdout,
    )
    writer.result(timeout=1)


def test_upload_file__hyphen_file_instead_of_stdin(b2_cli, bucket, tmpdir, monkeypatch):
    """Test upload_file will upload file named `-` instead of stdin by default"""
    # TODO remove this in v4
    assert b2.__version__ < '4', "`-` filename should not be supported in next major version of CLI"
    filename = 'stdin.txt'
    content = "I'm very rare creature, a file named '-'"
    monkeypatch.chdir(str(tmpdir))
    source_file = tmpdir.join('-')
    source_file.write(content)

    expected_stdout = f'URL by file name: http://download.example.com/file/my-bucket/{filename}'
    expected_json = {
        "action": "upload",
        "contentSha1": "ab467567b98216a255f77aef08aa2c418073d974",
        "fileName": filename,
        "size": len(content),
    }
    b2_cli.run(
        ['upload-file', '--no-progress', 'my-bucket', '-', filename],
        expected_json_in_stdout=expected_json,
        remove_version=True,
        expected_part_of_stdout=expected_stdout,
        expected_stderr=
        "WARNING: Filename `-` won't be supported in the future and will always be treated as stdin alias.\n",
    )


def test_upload_file__stdin(b2_cli, bucket, tmpdir, mock_stdin):
    """Test upload_file stdin alias support"""
    content = "stdin input"
    filename = 'stdin.txt'

    expected_stdout = f'URL by file name: http://download.example.com/file/my-bucket/{filename}'
    expected_json = {
        "action": "upload",
        "contentSha1": "2ce72aa159d1f190fddf295cc883f20c4787a751",
        "fileName": filename,
        "size": len(content),
    }
    mock_stdin.write(content)
    mock_stdin.close()

    b2_cli.run(
        ['upload-file', '--no-progress', 'my-bucket', '-', filename],
        expected_json_in_stdout=expected_json,
        remove_version=True,
        expected_part_of_stdout=expected_stdout,
    )


def test_upload_file__threads_setting(b2_cli, bucket, tmp_path):
    """Test upload_file supports setting number of threads"""
    num_threads = 66
    filename = 'file1.txt'
    content = 'hello world'
    local_file1 = tmp_path / 'file1.txt'
    local_file1.write_text(content)

    expected_json = {
        "action": "upload",
        "contentSha1": "2aae6c35c94fcfb415dbe95f408b9ce91ee846ed",
        "fileInfo": {
            "src_last_modified_millis": f"{local_file1.stat().st_mtime_ns // 1000000}"
        },
        "fileName": filename,
        "size": len(content),
    }

    b2_cli.run(
        [
            'upload-file', '--no-progress', 'my-bucket', '--threads',
            str(num_threads),
            str(local_file1), 'file1.txt'
        ],
        expected_json_in_stdout=expected_json,
        remove_version=True,
    )

    assert b2_cli.console_tool.api.services.upload_manager.get_thread_pool_size() == num_threads



######################################################################
#
# File: test/helpers.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import platform

import pytest

_MISSING = object()


def skip_on_windows(*args, reason='Not supported on Windows', **kwargs):
    return pytest.mark.skipif(
        platform.system() == 'Windows',
        reason=reason,
    )(*args, **kwargs)


def b2_uri_args_v3(bucket_name, path=_MISSING):
    if path is _MISSING:
        return [bucket_name]
    else:
        return [bucket_name, path]


def b2_uri_args_v4(bucket_name, path=_MISSING):
    if path is _MISSING:
        path = ''
    return [f'b2://{bucket_name}/{path}']



######################################################################
#
# File: b2/_internal/_cli/obj_dumps.py
#
# Copyright 2024 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import io

from b2sdk.v2 import (
    unprintable_to_hex,
)

_simple_repr_map = {
    False: "false",
    None: "null",
    True: "true",
}
_simple_repr_map_values = set(_simple_repr_map.values()) | {"yes", "no"}


def _yaml_simple_repr(obj):
    """
    Like YAML for simple types, but also escapes control characters for safety.
    """
    if isinstance(obj, (int, float)) and not isinstance(obj, bool):
        return str(obj)
    simple_repr = _simple_repr_map.get(obj)
    if simple_repr:
        return simple_repr
    obj_repr = unprintable_to_hex(str(obj))
    if isinstance(
        obj, str
    ) and (obj == "" or obj_repr.lower() in _simple_repr_map_values or obj_repr.isdigit()):
        obj_repr = repr(obj)  # add quotes to distinguish from numbers and booleans
    return obj_repr


def _id_name_first_key(item):
    try:
        return ("id", "name").index(str(item[0]).lower()), item[0], item[1]
    except ValueError:
        return 2, item[0], item[1]


def _dump(data, indent=0, skip=False, *, output):
    prefix = " " * indent
    if isinstance(data, dict):
        for idx, (key, value) in enumerate(sorted(data.items(), key=_id_name_first_key)):
            output.write(f"{'' if skip and idx == 0 else prefix}{_yaml_simple_repr(key)}: ")
            if isinstance(value, (dict, list)):
                output.write("\n")
                _dump(value, indent + 2, output=output)
            else:
                _dump(value, 0, True, output=output)
    elif isinstance(data, list):
        for idx, item in enumerate(data):
            output.write(f"{'' if skip and idx == 0 else prefix}- ")
            _dump(item, indent + 2, True, output=output)
    else:
        output.write(f"{'' if skip else prefix}{_yaml_simple_repr(data)}\n")


def readable_yaml_dump(data, output: io.TextIOBase) -> None:
    """
    Print YAML-like human-readable representation of the data.

    :param data: The data to be printed. Can be a list, dict, or any basic datatype.
    :param output: An output stream derived from io.TextIOBase where the data is to be printed.
    """
    _dump(data, output=output)



######################################################################
#
# File: test/unit/_cli/test_obj_dumps.py
#
# Copyright 2024 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from io import StringIO

import pytest

from b2._internal._cli.obj_dumps import readable_yaml_dump

# Test cases as tuples: (input_data, expected_output)
test_cases = [
    ({"key": "value"}, "key: value\n"),
    ([{"a": 1, "b": 2}], "- a: 1\n  b: 2\n"),
    ([1, 2, "false"], "- 1\n- 2\n- 'false'\n"),
    ({"true": True, "null": None}, "'null': null\n'true': true\n"),
    ([1., 0.567], "- 1.0\n- 0.567\n"),
    ([''], "- ''\n"),
    (
        # make sure id and name are first, rest should be sorted alphabetically
        [
            {"b": 2, "a": 1, "name": 4, "id": 3},
        ],
        "- id: 3\n  name: 4\n  a: 1\n  b: 2\n",
    ),
    (  # nested data
        [
            {
                "name": "John Doe",
                "age": 30,
                "addresses": [
                    {
                        "street": "123 Elm St",
                        "city": "Somewhere",
                    },
                    {
                        "street": "456 Oak St",
                    },
                ],
                "address": {
                    "street": "789 Pine St",
                    "city": "Anywhere",
                    "zip": "67890",
                },
            }
        ],
        (
            "- name: John Doe\n"
            "  address: \n"
            "    city: Anywhere\n"
            "    street: 789 Pine St\n"
            "    zip: '67890'\n"
            "  addresses: \n"
            "    - city: Somewhere\n"
            "      street: 123 Elm St\n"
            "    - street: 456 Oak St\n"
            "  age: 30\n"
        ),
    ),
]


@pytest.mark.parametrize("input_data,expected", test_cases)
def test_readable_yaml_dump(input_data, expected):
    output = StringIO()
    readable_yaml_dump(input_data, output)
    assert output.getvalue() == expected
